---------------------------------------------------------------
--                                                           
--  SCRIPT PROGRAM    
--                                                           
--  SCRIPT.ADB 
--  Description : runs another executable, echoing all I/O
--                to a specified file
--                from the Run menu:
--  Usage:
--    script <output_filename> <command_name> [command_arguments]
--
--  By: Dr. Martin C. Carlisle
--      US Air Force Academy
--                                         
-- SCRIPT is free software; you can redistribute it and/or 
-- modify it without restriction.  However, we ask that you
-- please retain the original author information, and clearly  
-- indicate if it has been modified.
--
-- SCRIPT is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
--
---------------------------------------------------------------
-- LIMITATIONS:  Will not work properly with programs that use
--               Get_Immediate:  requires extra presses of <CR>
--
---------------------------------------------------------------
-- Change log:
--   06/19/98 (mcc) : added isatty test
--   12/17/97 (mcc) : now do both PeekNamedPipe and PeekConsoleInput
--                    as with 3.11a, PeekConsoleInput doesn't work
--                    if input is a pipe
--   09/10/97 (mcc) : added ReadAllInput after process completed
--   08/01/97 (mcc) : added check after peek to ensure is a keydown
--------------------------------------

with Win32.Winbase,Win32.Winuser,Win32.Winnt,Win32.Wincon;
with System;
with Winconversions;
with Ada.Command_Line,Ada.Characters.Handling;
with Ada.Text_Io,Ada.Unchecked_Conversion;
use type Win32.Dword,Win32.Bool,Win32.Winnt.Handle,Win32.Word;
with Ada.Strings.Bounded;
with Interfaces.C_Streams;

procedure Script is
   -- Uncomment following to compile with GNAT
   pragma Linker_Options("c:\adagide\win32ada\win32ada.a");
   pragma Linker_Options("c:\usr\lib\libadvapi32.a");
   pragma Linker_Options("c:\usr\lib\libuser32.a");

   package Argumentstringpkg is new
      Ada.Strings.Bounded.Generic_Bounded_Length(255);

   procedure Usage is
   begin
      Ada.Text_Io.Put("Your command was: " & Ada.Command_Line.Command_Name);
      for I in 1..Ada.Command_Line.Argument_Count loop
         Ada.Text_Io.Put(" " & Ada.Command_Line.Argument(I));
      end loop;
      Ada.Text_Io.New_Line;

      Ada.Text_Io.Put_Line("Usage: " & Ada.Command_Line.Command_Name &
         " output_filename executable [command_arguments]");
      Ada.Text_Io.New_Line;
      Ada.Text_Io.Put_Line("Press <enter> to exit");
      Ada.Text_Io.Skip_Line;
   end Usage;

   function Convert is new Ada.Unchecked_Conversion(Win32.Lpstr,
      Win32.Lpvoid);

   Si                  : aliased Win32.Winbase.Startupinfo;
   Pi                  : aliased Win32.Winbase.Process_Information;
   Sa                  : aliased Win32.Winbase.Security_Attributes;
   Created             : Win32.Bool;
   Startpath           : array(0..255) of aliased Win32.Char;
   Buffer              : array(0..1023) of aliased Win32.Char;
   Inputbuffer         : aliased Win32.Wincon.Input_Record;
   Ignoredword         : Win32.Dword;
   Dwordresult         : aliased Win32.Dword;
   Ignorebool          : Win32.Bool;
   Stuffinpipe         : Win32.Bool;
   Howmanybytes        : aliased Win32.Dword;
   Readargument        : aliased Win32.Dword;
   Argumentstring      : Argumentstringpkg.Bounded_String :=
      Argumentstringpkg.To_Bounded_String(Source => "");
   Argumentcounter     : Integer := 1;
   Promptforarguments  : Boolean := False;
   Amscripting         : Boolean := False;
   Scriptargument      : constant Integer := 1;
   Executableargument  : constant Integer := 2;
   Pipereadinput       : aliased Win32.Winnt.Handle;
   Pipewriteinput      : aliased Win32.Winnt.Handle;
   Pipereadoutput      : aliased Win32.Winnt.Handle;
   Pipewriteoutput     : aliased Win32.Winnt.Handle;
   Standardinput       : Win32.Winnt.Handle;
   Standardoutput      : Win32.Winnt.Handle;
   Outputfile          : Win32.Winnt.Handle;
   Redirected          : Boolean := False;

   procedure Readallinput is
   begin
      loop
         Stuffinpipe := Win32.Winbase.Peeknamedpipe(Pipereadoutput,
            System.Null_Address, 0, -- don't actually read anything
            null, -- don't care
            Howmanybytes'Unchecked_Access,
            null);

         exit when Stuffinpipe = 0 or else Howmanybytes = 0;

         if Howmanybytes > Win32.Dword(Buffer'Length) then
            Readargument := Win32.Dword(Buffer'Length);
         else
            Readargument := Howmanybytes;
         end if;
         Ignorebool:=Win32.Winbase.Readfile(Pipereadoutput,
            Convert(Buffer(Buffer'First)'Unchecked_Access),
            Readargument,
            Howmanybytes'Unchecked_Access,null);
         Ignorebool:= Win32.Winbase.Writefile(Standardoutput,
            Convert(Buffer(Buffer'First)'Unchecked_Access),
            Howmanybytes,
            Readargument'Unchecked_Access,null);
         Ignorebool:= Win32.Winbase.Writefile(Outputfile,
            Convert(Buffer(Buffer'First)'Unchecked_Access),
            Howmanybytes,
            Readargument'Unchecked_Access,null);
      end loop;
   end Readallinput;

   procedure Read_Keyboard_And_Echo is
      Ignorebool : Win32.Bool;
   begin
      -- uncomment for debugging
      --      Ada.Text_IO.Put_Line("start readfile");

      Ignorebool := Win32.Winbase.Readfile(Standardinput,
         Convert(Buffer(Buffer'First)'Unchecked_Access),
         Win32.Dword(Buffer'Length),
         Howmanybytes'Unchecked_Access,null);

      -- uncomment for debugging
      --      Ada.Text_Io.Put_Line("writing output to pipe" &
      --         Integer'Image(Integer(Howmanybytes)));

      Ignorebool:= Win32.Winbase.Writefile(Pipewriteinput,
         Convert(Buffer(Buffer'First)'Unchecked_Access),
         Howmanybytes,
         Readargument'Unchecked_Access,null);
      Ignorebool:= Win32.Winbase.Flushfilebuffers(Pipewriteinput);
      Ignorebool:= Win32.Winbase.Writefile(Outputfile,
         Convert(Buffer(Buffer'First)'Unchecked_Access),
         Howmanybytes,
         Readargument'Unchecked_Access,null);
   end Read_Keyboard_And_Echo;

   -- START of MAIN program
begin

   if Ada.Command_Line.Argument_Count < 2 then
      Usage;
      return;
   end if;

   if Interfaces.C_Streams.Isatty(
      Interfaces.C_Streams.Fileno(
         Interfaces.C_Streams.Stdin)) = 0 then
      Redirected := True;
   end if;

   for I in Executableargument+1..Ada.Command_Line.Argument_Count loop
      Argumentstring := Argumentstringpkg."&"(Argumentstring,
         " " & Ada.Command_Line.Argument(I));
   end loop;

   Sa.Nlength := Win32.Dword(Sa'Size/8);
   Sa.Lpsecuritydescriptor := System.Null_Address;
   Sa.Binherithandle := 1; -- Win32.BOOL(TRUE)

   if not Redirected then
      Created := Win32.Winbase.Createpipe(Pipereadinput'Unchecked_Access,
         Pipewriteinput'Unchecked_Access,Sa'Unchecked_Access,
         Win32.Dword(1024));
      if Created = 0 then
         Ada.Text_Io.Put_Line("Internal error: cannot create input pipe"&
            Standard.Ascii.Lf & Standard.Ascii.Lf &
            "Error code:" & Integer'Image(Integer(Win32.Winbase.Getlasterror)));
         return;
      end if;
   end if;

   Created := Win32.Winbase.Createpipe(Pipereadoutput'Unchecked_Access,
      Pipewriteoutput'Unchecked_Access,Sa'Unchecked_Access,
      Win32.Dword(1024));
   if Created = 0 then
      Ada.Text_Io.Put_Line("Internal error: cannot create output pipe"&
         Standard.Ascii.Lf & Standard.Ascii.Lf &
         "Error code:" & Integer'Image(Integer(Win32.Winbase.Getlasterror)));
      return;
   end if;

   Si.Cb := Win32.Dword(Si'Size/8);
   Si.Lpreserved := null;
   Si.Lpdesktop := null;
   Si.Lptitle := null;
   Si.Cbreserved2 := 0;
   Si.Dwflags := Win32.Dword(Win32.Winbase.Startf_Useshowwindow +
      Win32.Winbase.Startf_Usestdhandles);
   Si.Wshowwindow := Win32.Word(Win32.Winuser.Sw_Normal);
   Si.Lpreserved2 := null;
   if not Redirected then
      Si.Hstdinput := Pipereadinput;
   else
      Si.Hstdinput := Win32.Winbase.Getstdhandle(
         Win32.Winbase.Std_Input_Handle);
   end if;
   Si.Hstdoutput := Pipewriteoutput;
   Si.Hstderror := Pipewriteoutput;

   Standardinput := Win32.Winbase.Getstdhandle(Win32.Winbase.Std_Input_Handle);
   Standardoutput := Win32.Winbase.Getstdhandle(Win32.Winbase.Std_Output_Handle);


   Ignoredword := Win32.Winbase.Getcurrentdirectory(
      Win32.Dword(Startpath'Length),
      Startpath(Startpath'First)'Unchecked_Access);

   declare
      Command_Name         : String := Ada.Command_Line.Command_Name;
      Stringcounter        : Integer;
      Scriptargumentstring : constant String :=
         Ada.Command_Line.Argument(Scriptargument) & Character'First;
      Executablenamestring : String :=
         Ada.Command_Line.Argument(Executableargument) & ".exe";
      Executablenamelength : Integer;
      Fileptr              : Ada.Text_Io.File_Type;
   begin
      Sa.Nlength := Win32.Dword(Sa'Size/8);
      Sa.Lpsecuritydescriptor := System.Null_Address;
      Sa.Binherithandle := 1; -- Win32.BOOL(TRUE)

      -- we need to create the file first, so that it exists and
      Outputfile := Win32.Winbase.Createfile(
         Winconversions.Convertstaticstring(Scriptargumentstring),
         Win32.Winnt.Generic_Write,
         0,
         Sa'Unchecked_Access,
         Win32.Winbase.Create_Always,
         Win32.Winnt.File_Attribute_Normal,System.Null_Address);

      if Outputfile = Win32.Winbase.Invalid_Handle_Value then
         Ada.Text_Io.Put_Line("Unable to open output file: " &
            Scriptargumentstring);
      end if;

      -- we appended .exe to executable name, if it was already present,
      -- move null terminator forward.
      if Ada.Characters.Handling.To_Lower(Executablenamestring(
               Executablenamestring'Last-4)) = 'e' and then
            Ada.Characters.Handling.To_Lower(Executablenamestring(
               Executablenamestring'Last-5)) = 'x' and then
            Ada.Characters.Handling.To_Lower(Executablenamestring(
               Executablenamestring'Last-6)) = 'e' and then
            Executablenamestring(Executablenamestring'Last-7) = '.' then
         Executablenamelength := Executablenamestring'Last-4;
      else
         Executablenamelength := Executablenamestring'Last;
      end if;

      Stringcounter := Command_Name'Last;
      -- find end of path
      while Stringcounter >= Command_Name'First and then
            Command_Name(Stringcounter) /= '\' loop
         Stringcounter := Stringcounter - 1;
      end loop;

      -- if no path, simply stick a ".\" at the front
      -- to avoid exceptions/complications in declare block below
      if Stringcounter < Command_Name'First then
         Command_Name(Command_Name'First)   := '.';
         Command_Name(Command_Name'First+1) := '\';
         Stringcounter := Command_Name'First+1;
      end if;


      declare
         Script_Name : constant String :=
            Executablenamestring(Executablenamestring'First..
            Executablenamelength) &
            Argumentstringpkg.To_String(Argumentstring) & Character'First;
      begin
         Ada.Text_Io.Put_Line("Script v 1.4 Executing: " & Script_Name);
         Ada.Text_Io.Put_Line("Output to file: " & Scriptargumentstring);
         Created := Win32.Winbase.Createprocess(
            null,
            Winconversions.Convertstaticstring(Script_Name),
            null,
            null,
            Win32.Bool(1),  -- inherit handles
            Win32.Dword(0), -- flags
            System.Null_Address,
            Startpath(0)'Unchecked_Access,
            Si'Unchecked_Access,
            Pi'Unchecked_Access);
         if Created = 0 or Pi.Hprocess = System.Null_Address then
            Ada.Text_Io.Put_Line("Unable to start process");
            Ada.Text_Io.Put_Line("Error code:" &
               Integer'Image(Integer(Win32.Winbase.Getlasterror)));
         end if;
      end; -- inner declare block
   end; -- outer declare block


   Ignorebool := Win32.Winbase.Closehandle(Pipewriteoutput);
   Ignorebool := Win32.Winbase.Closehandle(Pipereadinput);

   -- from Win32 programming bible book 1 p. 1208
   Ignoredword := Win32.Winuser.Waitforinputidle(Win32.Winbase.Getcurrentprocess,
      Win32.Winbase.Infinite);


   if Pi.Hprocess /= System.Null_Address then
      Dwordresult := Win32.Winbase.Still_Active;

      while Dwordresult = Win32.Winbase.Still_Active loop
         -- read all possible input first
         Readallinput;

         -- first try to read as a pipe
         if not Redirected then
            Ignorebool := Win32.Winbase.Peeknamedpipe(Standardinput,
               Buffer(Buffer'First)'Address,
               Win32.Dword(Buffer'Length),
               Howmanybytes'Unchecked_Access,
               Readargument'Unchecked_Access,
               Dwordresult'Unchecked_Access);
            -- if something present, echo
            if Integer(Ignorebool) /= 0 and then
                  Integer(Howmanybytes) > 0 then
               -- uncomment to debug
               --   Ada.Text_IO.Put_Line("peeked as pipe");
               Read_Keyboard_And_Echo;
            end if;

            -- then try to read as a console
            Ignorebool := Win32.Wincon.Peekconsoleinput(
               Standardinput,
               Inputbuffer'Unchecked_Access,
               Win32.Dword(1),
               Howmanybytes'Unchecked_Access);
            -- if something present, echo
            if Integer(Howmanybytes) = 1 then
               if Inputbuffer.Eventtype = Win32.Wincon.Key_Event and then
                     Win32.Wincon.Key_Event_Record(Inputbuffer.Event.Keyevent).Bkeydown /= 0 then
                  Read_Keyboard_And_Echo;
               else
                  Ignorebool := Win32.Wincon.Readconsoleinput(
                     Standardinput,
                     Inputbuffer'Unchecked_Access,
                     Win32.Dword(1),
                     Howmanybytes'Unchecked_Access);
               end if;
            end if;
         end if;
         Ignorebool := Win32.Winbase.Getexitcodeprocess(Pi.Hprocess,
            Dwordresult'Unchecked_Access);
      end loop;

      Readallinput; -- make sure no extraneous stuff left in buffer
   end if;

   if not Redirected then
      Ignorebool := Win32.Winbase.Closehandle(Pipewriteinput);
   end if;
   Ignorebool := Win32.Winbase.Closehandle(Pipereadoutput);
   Ignorebool := Win32.Winbase.Closehandle(Outputfile);

end Script;